;;; pcap.scm v0.5 - 31 March 2004 - Category 5
;;;
;;; An interface to the libpcap portable packet-capture library, for
;;; the Chicken Scheme compiler.

; ----------------------------TERMS OF USE--------------------------------
; Copyright (c) 2003-2004, Category 5
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are
; met:
;
;   Redistributions of source code must retain the above copyright notice,
;   this list of conditions and the following disclaimer. Redistributions in
;   binary form must reproduce the above copyright notice, this list of
;   conditions and the following disclaimer in the documentation and/or
;   other materials provided with the distribution. Neither the name of the
;   author nor the names of its contributors may be used to endorse or
;   promote products derived from this software without specific prior
;   written permission.
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
; IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
; THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
; PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
; EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
; PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
; PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
; LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
; ----------------------------TERMS OF USE--------------------------------


(declare
 (uses lolevel)
 (usual-integrations)
 (fixnum-arithmetic)
 (no-bound-checks)
 (export pcap-open-live pcap-lookupnet pcap-lookupdev pcap-loop pcap-next
         pcap-open-offline pcap-dump-open pcap-dump-close
         pcap-close pcap-geterr pcap-compile pcap-setfilter pcap-stats
         pcap-datalink pcap-snapshot pcap-swapped? pcap-major-version
         pcap-dispatch pcap-loop/dump pcap-dispatch/dump pcap-minor-version))
; (bound-to-procedure))
  
(register-feature! 'pcap)

(cond-expand
 [unsafe
  (eval-when (compile)
             (define-macro (##sys#check-structure x y) '(##core#undefined))
             (define-macro (##sys#check-range x y z) '(##core#undefined))
             (define-macro (##sys#check-pair x) '(##core#undefined))
             (define-macro (##sys#check-list x) '(##core#undefined))
             (define-macro (##sys#check-symbol x) '(##core#undefined))
             (define-macro (##sys#check-string x) '(##core#undefined))
             (define-macro (##sys#check-char x) '(##core#undefined))
             (define-macro (##sys#check-exact x) '(##core#undefined))
             (define-macro (##sys#check-port x) '(##core#undefined))
             (define-macro (##sys#check-number x) '(##core#undefined))
             (define-macro (##sys#check-byte-vector x) '(##core#undefined)))]
 [else])


#>
#include <pcap.h>
<#

(define-foreign-variable bpf_program_size unsigned-int "sizeof(struct bpf_program)")

(define pointer->string
  (lambda (p len)
    (let ((s (make-string len)))
      (move-memory! p s len)
      s)))
(define protected-pointer->object
  (foreign-lambda* scheme-object (((pointer (pointer "C_word")) ptr))
    "return(**ptr);"))

(define-external (scheme_packet_callback
                  ((c-pointer unsigned-char) userdata)
                  ((const (c-pointer (struct "pcap_pkthdr"))) header)
                  ((const (c-pointer unsigned-char)) data))
  void
  (let ((h (%pkthdr-unpack header))
        (p (protected-pointer->object userdata)))
    (assert (procedure? p) "(assertion failed in scheme_packet_callback)" p)
    (apply p (pointer->string data (car h)) h))) ; (p data len secs usecs)

(define %pkthdr-unpack
  (foreign-safe-lambda*
   scheme-object
   (((c-pointer (struct "pcap_pkthdr")) h))
   #<<EOF
     C_word *ptr = C_alloc(C_SIZEOF_LIST(3));
     C_word lst;
     lst = C_list(&ptr, 3, C_fix(h->len), C_fix(h->ts.tv_sec), C_fix(h->ts.tv_usec));
     return(lst);
EOF
))

(define pcap-loop
  (foreign-safe-lambda*
     int
     (((c-pointer (struct "pcap")) p)
      (int cnt) (scheme-object c))
#<<EOF
     int result;
     C_word *data[1];

     data[0] = &c;
     C_gc_protect(data, 1);
     result = pcap_loop(p, cnt, scheme_packet_callback, (unsigned char *)data);
     C_gc_unprotect(1);
     
     return(result);
EOF
))
(define pcap-loop/dump
  (foreign-safe-lambda*
     int
     (((c-pointer (struct "pcap")) p) (int cnt))
#<<EOF
     return(pcap_loop(p, cnt, pcap_dump, NULL));
EOF
))

(define pcap-dispatch
  (foreign-safe-lambda*
     int
     (((c-pointer (struct "pcap")) p)
      (int cnt) (scheme-object c))
#<<EOF
     int result;
     C_word *data[1];

     data[0] = &c;
     C_gc_protect(data, 1);
     result = pcap_dispatch(p, cnt, scheme_packet_callback, (unsigned char *)data);
     C_gc_unprotect(1);
     
     return(result);
EOF
))
(define pcap-dispatch/dump
  (foreign-safe-lambda*
     int
     (((c-pointer (struct "pcap")) p) (int cnt))
#<<EOF
     return(pcap_dispatch(p, cnt, pcap_dump, NULL));
EOF
))

(define-foreign-variable PCAP_ERRBUF_SIZE int)
(define errbuf (make-string PCAP_ERRBUF_SIZE))

(define %pcap-lookupdev
  (foreign-lambda c-string pcap_lookupdev pointer) )
(define pcap-lookupdev
  (lambda ()
    (%pcap-lookupdev errbuf)))

(define %pcap-lookupnet
  (foreign-lambda int pcap_lookupnet
                  c-string (pointer unsigned-int) (pointer unsigned-int)
                  pointer))
(define pcap-lookupnet
  (lambda (dev)
    (let-location ((netp unsigned-int 0)
                   (maskp unsigned-int 0) )
      (let ((ret (%pcap-lookupnet dev
                                  (location netp)
                                  (location maskp)
                                  errbuf)))
        (if (eq? ret -1)
            (values #f #f)
            (values (pointer-u32-ref (location netp))
                    (pointer-u32-ref (location maskp))))))))

(define pcap-open-live
  (foreign-lambda (c-pointer (struct "pcap")) pcap_open_live
                  c-string int bool int c-string))
(define pcap-open-offline
  (foreign-lambda (c-pointer (struct "pcap")) pcap_open_offline
                  c-string (c-pointer char)))
(define pcap-dump-open
  (foreign-lambda (c-pointer (struct "pcap_dumper")) pcap_dump_open
                  (c-pointer (struct "pcap")) c-string))
; (define pcap-dump
;   (foreign-lambda void pcap_dump
;                   (c-pointer unsigned-char)
;                   (c-pointer (struct "pcap_pkthdr"))
;                   (c-pointer unsigned-char)))
(define pcap-dump-close
  (foreign-lambda void pcap_dump_close (c-pointer (struct "pcap_dumper"))))
(define pcap-close
  (foreign-lambda void pcap_close (pointer (struct "pcap"))))

(define pcap-next
  (lambda (p)
    (apply values (%pcap-next p))))
(define-external (pcap_next_callback ((c-pointer unsigned-char) data)
                                     (int len) (int sec) (int usec))
  scheme-object
  (let ((buf (make-string len)))
    (move-memory! data buf len)
    (list buf sec usec)))
(define %pcap-next
    (foreign-safe-lambda*
     scheme-object (((c-pointer (struct "pcap")) p))
#<<EOF
     static struct pcap_pkthdr ph;
     const unsigned char *data;
     data = pcap_next(p, &ph);
     if (data == NULL)
       return(C_SCHEME_FALSE);
     return(pcap_next_callback((unsigned char *)data,
                               ph.len, ph.ts.tv_sec, ph.ts.tv_usec));
EOF
))

(define pcap-geterr
  (foreign-lambda c-string pcap_geterr (c-pointer (struct "pcap"))))
(define %pcap-compile
  (foreign-lambda int pcap_compile
                  (c-pointer (struct "pcap"))
                  (c-pointer (struct "bpf_program"))
                  c-string
                  int
                  unsigned-int))
(define pcap-compile
  (lambda (handle str netmask)
    (let* ((bpf-program (allocate bpf_program_size))
           (result (%pcap-compile handle bpf-program str 1 netmask)))
      (if (zero? result) bpf-program #f))))
(define pcap-setfilter
  (foreign-lambda int pcap_setfilter
                  (c-pointer (struct "pcap"))
                  (c-pointer (struct "bpf_program"))))

(define pcap-stats
  (lambda (p)
    (apply values (%pcap-stats p))))
(define %pcap-stats
  (foreign-lambda* scheme-object
                   (((c-pointer (struct "pcap")) p))
#<<EOF
     struct pcap_stat ps;
     C_word *ptr = C_alloc(C_SIZEOF_LIST(2));
     C_word lst;

     if (pcap_stats(p, &ps) == -1)
       return(C_SCHEME_FALSE);
     lst = C_list(&ptr, 2, C_fix(ps.ps_recv), C_fix(ps.ps_drop));
     return(lst);
EOF
))
(define pcap-datalink
  (foreign-lambda int pcap_datalink (c-pointer (struct "pcap"))))
(define pcap-snapshot
  (foreign-lambda int pcap_snapshot (c-pointer (struct "pcap"))))
(define pcap-swapped?
  (foreign-lambda bool pcap_is_swapped (c-pointer (struct "pcap"))))
(define pcap-major-version
  (foreign-lambda int pcap_major_version (c-pointer (struct "pcap"))))
(define pcap-minor-version
  (foreign-lambda int pcap_minor_version (c-pointer (struct "pcap"))))
